@@ -62,7 +62,7 @@ failed_store:
 }
 
 /* we cannot log in at all, notify clients */
-void kill_pool_logins(PgPool *pool, const char *msg)
+void kill_pool_logins(PgPool *pool, const char *sqlstate, const char *msg)
 {
 	struct List *item, *tmp;
 	PgSocket *client;
@@ -72,18 +72,19 @@ void kill_pool_logins(PgPool *pool, const char *msg)
 		if (!client->wait_for_welcome)
 			continue;
 
-		disconnect_client(client, true, "%s", msg);
+		disconnect_client_sqlstate(client, true, sqlstate, msg);
 	}
 }
 
 /* we cannot log in at all, notify clients with server error */
 static void kill_pool_logins_server_error(PgPool *pool, PktHdr *errpkt)
 {
-	const char *level, *msg;
+	const char *level, *msg, *sqlstate;
 
-	parse_server_error(errpkt, &level, &msg);
+	parse_server_error(errpkt, &level, &msg, &sqlstate);
 	log_warning("server login failed: %s %s", level, msg);
-	kill_pool_logins(pool, msg);
+	log_noise("kill_pool_logins_server_error: sqlstate: %s", sqlstate);
+	kill_pool_logins(pool, sqlstate, msg);
 }
 
 /* process packets on server auth phase */
@@ -455,6 +456,9 @@ static bool handle_connect(PgSocket *server)
 	 * Only if there are no cancel requests we proceed with the login procedure
 	 * that's necessary to handle queries. Cancel requests need to be sent
 	 * before the login procedure starts.
+	 *
+	 * A special case is when this is a peer pool, instead of a regular pool.
+	 * Since only cancellation requests should be sent to peers.
 	 */
 	if (!statlist_empty(&pool->waiting_cancel_req_list)) {
 		slog_debug(server, "use it for pending cancel req");
@@ -466,7 +470,11 @@ static bool handle_connect(PgSocket *server)
 			server->ready = true;
 			disconnect_server(server, false, "failed to send cancel req");
 		}
-	} else {
+	} else if(pool->db->peer_id) {
+		/* notify disconnect_server() that connect did not fail */
+		server->ready = true;
+		disconnect_server(server, false, "peer server was not necessary anymore, because client cancel connection was already closed");
+	}else {
 		/* proceed with login */
 		if (server_connect_sslmode > SSLMODE_DISABLED && !is_unix) {
 			slog_noise(server, "P: SSL request");
@@ -576,6 +584,7 @@ bool server_proto(SBuf *sbuf, SBufEvent evtype, struct MBuf *data)
 		case SV_TESTED:
 		case SV_USED:
 		case SV_ACTIVE:
+		case SV_ACTIVE_CANCEL:
 		case SV_IDLE:
 			res = handle_server_work(server, &pkt);
 			break;
@@ -611,6 +620,7 @@ bool server_proto(SBuf *sbuf, SBufEvent evtype, struct MBuf *data)
 			server->resetting = false;
 			switch (server->state) {
 			case SV_ACTIVE:
+			case SV_ACTIVE_CANCEL:
 			case SV_TESTED:
 				/* keep link if client expects more Syncs */
 				if (server->link) {
